(define-module (ordo connection)
  #:use-module (oop goops)
  #:use-module (ice-9 exceptions)
  #:use-module (ice-9 match)
  #:use-module (ice-9 popen)
  #:use-module (ice-9 rdelim)
  #:use-module (logging logger)
  #:use-module (srfi srfi-1)  ; list operations
  #:use-module (srfi srfi-71) ; extended let
  #:use-module (ordo connection base)
  #:use-module (ordo connection local)
  #:use-module (ordo connection ssh)
  #:use-module (ordo connection sudo)
  #:use-module (ordo util flatten)
  #:use-module (ordo util shell-quote)
  #:use-module (ordo util keyword-args)
  #:export (connection
            call-with-connection
            run))

(define (connection type . kwargs)
  (validate-keyword-args kwargs)
  (let* ((c (case type
              ((#:local) (make <local-connection>))
              ((#:ssh) (apply make <ssh-connection>
                              (select-keyword-args kwargs '(#:user #:host #:password #:identity #:authenticate-server?))))))
         (c (if (keyword-arg kwargs #:sudo?)
                (apply make <sudo-connection> #:connection c (select-keyword-args kwargs '(#:become-user #:become-password)))
                c)))
    (conn:validate c)
    c))

(define (call-with-connection c proc)
  (dynamic-wind
    (lambda () (conn:setup c))
    (lambda () (proc c))
    (lambda () (conn:teardown c))))

(define (build-command prog args pwd env)
  (let ((xs (remove unspecified?
                    (flatten (list "env"
                                   (when pwd (list "--chdir" (string-shell-quote pwd)))
                                   (when env (map (match-lambda ((k . v) (string-append k "=" (string-shell-quote v)))) env))
                                   prog
                                   (map string-shell-quote args)
                                   "2>&1")))))
    (string-join xs " ")))

(define (run conn prog . args)
  (let* ((args (flatten args))
         (args kwargs (break keyword? args))
         (args (remove unspecified? args))
         (pwd (keyword-arg kwargs #:pwd))
         (env (keyword-arg kwargs #:env))
         (return (keyword-arg kwargs #:return identity))
         (check? (keyword-arg kwargs #:check?))
         (command (build-command prog args pwd env))
         (out rc (conn:run conn command)))
    (log-msg 'INFO "Command " command " exited " rc)
    (if check?
        (if (zero? rc)
            (return out)
            (raise-exception (make-exception
                              (make-external-error)
                              (make-exception-with-message (format #f "Non-zero exit (~a) from ~a" rc prog)))))
        (values (return out) rc))))
